# This code is part of KQCircuits
# Copyright (C) 2024 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).

import logging
from typing import Callable
from kqcircuits.pya_resolver import pya
from kqcircuits.simulations.epr.utils import extract_child_simulation
from kqcircuits.simulations.partition_region import PartitionRegion
from kqcircuits.simulations.simulation import Simulation


# Partition region and correction cuts definitions for Swissmon qubit


def partition_regions(simulation: Simulation, prefix: str = "") -> list[PartitionRegion]:
    metal_edge_dimension = 4.0
    metal_edge_margin = pya.DPoint(metal_edge_dimension, metal_edge_dimension)
    cross_poly = pya.DPolygon([simulation.refpoints[f"epr_cross_{idx:02d}"] for idx in range(12)]).sized(
        metal_edge_dimension + simulation.island_r
    )

    result = [
        PartitionRegion(
            name=f"{prefix}crossmer",
            face=simulation.face_ids[0],
            metal_edge_dimensions=metal_edge_dimension,
            region=cross_poly,
            vertical_dimensions=3.0,
            visualise=True,
        ),
        PartitionRegion(
            name=f"{prefix}crossbulk",
            face=simulation.face_ids[0],
            metal_edge_dimensions=None,
            region=cross_poly,
            vertical_dimensions=3.0,
            visualise=True,
        ),
    ]

    # These are added to include the waveguides attached to couplers in the coupler region
    def _get_coupler_wg_offset(i, s):
        coupler_wg_offsets = [{"min": pya.DPoint(-150, 0)}, {"max": pya.DPoint(0, 150)}, {"max": pya.DPoint(150, 0)}]
        return coupler_wg_offsets[i].get(s, pya.DPoint(0, 0))

    for idx in range(3):
        # Need to check if coupler is present
        if simulation.cpl_length[idx] > 0:
            cplr_region = pya.DBox(
                simulation.refpoints[f"epr_cplr{idx}_min"] - metal_edge_margin + _get_coupler_wg_offset(idx, "min"),
                simulation.refpoints[f"epr_cplr{idx}_max"] + metal_edge_margin + _get_coupler_wg_offset(idx, "max"),
            )
            result += [
                PartitionRegion(
                    name=f"{prefix}{idx}cplrmer",
                    face=simulation.face_ids[0],
                    metal_edge_dimensions=metal_edge_dimension,
                    region=cplr_region,
                    vertical_dimensions=3.0,
                    visualise=True,
                ),
                PartitionRegion(
                    name=f"{prefix}{idx}cplrbulk",
                    face=simulation.face_ids[0],
                    metal_edge_dimensions=None,
                    region=cplr_region,
                    vertical_dimensions=3.0,
                    visualise=True,
                ),
            ]
    return result


def correction_cuts(simulation: Simulation, prefix: str = "") -> dict[str, dict]:
    cross_corner = simulation.refpoints["epr_cross_09"]
    coupler_corner = (
        simulation.refpoints["epr_cplr0_max"] if simulation.cpl_length[0] > 0 else simulation.refpoints["epr_cross_08"]
    )

    half_gap = simulation.gap_width[1] / 2
    if len(set(simulation.gap_width)) > 1:
        logging.warning("Partition regions for Swissmon with varying gaps are not implemented")
        logging.warning(
            "Using correction with %s gap for all arms with gap widths %s", str(2 * half_gap), str(simulation.gap_width)
        )

    cross_xsection_center = pya.DPoint((cross_corner.x + coupler_corner.x) / 2, cross_corner.y - half_gap)
    half_cut_length = 30.0
    result = {
        f"{prefix}crossmer": {
            "p1": cross_xsection_center + pya.DPoint(0, -half_cut_length),
            "p2": cross_xsection_center + pya.DPoint(0, half_cut_length),
            "metal_edges": [{"x": half_cut_length - half_gap}, {"x": half_cut_length + half_gap, "x_reversed": True}],
        }
    }

    half_gap = simulation.b / 2
    if simulation.cpl_length[0] > 0:
        xsection_point = simulation.cpl_gap[0] / 2 + simulation.cpl_width[0] / 2
        result[f"{prefix}0cplrmer"] = {
            "p1": simulation.refpoints["port_cplr0"] + pya.DPoint(-half_cut_length + half_gap, xsection_point),
            "p2": simulation.refpoints["port_cplr0"] + pya.DPoint(half_cut_length + half_gap, xsection_point),
            "metal_edges": [{"x": half_cut_length - half_gap}, {"x": half_cut_length + half_gap, "x_reversed": True}],
        }
    if simulation.cpl_length[1] > 0:
        xsection_point = simulation.cpl_gap[1] / 2 + simulation.cpl_width[1] / 2
        result[f"{prefix}1cplrmer"] = {
            "p1": simulation.refpoints["port_cplr1"] + pya.DPoint(xsection_point, half_cut_length - half_gap),
            "p2": simulation.refpoints["port_cplr1"] + pya.DPoint(xsection_point, -half_cut_length - half_gap),
            "metal_edges": [{"x": half_cut_length - half_gap}, {"x": half_cut_length + half_gap, "x_reversed": True}],
        }
    if simulation.cpl_length[2] > 0:
        xsection_point = simulation.cpl_gap[2] / 2 + simulation.cpl_width[2] / 2
        result[f"{prefix}2cplrmer"] = {
            "p1": simulation.refpoints["port_cplr2"] + pya.DPoint(-half_cut_length - half_gap, xsection_point),
            "p2": simulation.refpoints["port_cplr2"] + pya.DPoint(half_cut_length - half_gap, xsection_point),
            "metal_edges": [{"x": half_cut_length - half_gap}, {"x": half_cut_length + half_gap, "x_reversed": True}],
        }
    return result


def extract_swissmon_from(
    simulation: Simulation, refpoint_prefix: str, parameter_remap_function: Callable[[Simulation, str], any]
):
    return extract_child_simulation(
        simulation,
        refpoint_prefix,
        parameter_remap_function,
        [
            "b",
            "gap_width",
            "face_ids",  # Accesses list's index 0
            "island_r",
            "cpl_gap",  # Accesses list's indices 0, 1, 2
            "cpl_width",  # Accesses list's indices 0, 1, 2
            "cpl_length",  # Accesses list's indices 0, 1, 2
        ],
    )



# This code is part of KQCircuits
# Copyright (C) 2024 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).
from pathlib import Path

from typing import Callable, Sequence

from kqcircuits.simulations.simulation import Simulation

from kqcircuits.simulations.partition_region import get_list_of_two

from kqcircuits.simulations.export.simulation_export import cross_combine

from kqcircuits.simulations.export.elmer.elmer_solution import ElmerCrossSectionSolution

from kqcircuits.defaults import XSECTION_PROCESS_PATH

from kqcircuits.simulations.export.xsection.xsection_export import (
    separate_signal_layer_shapes,
    create_xsections_from_simulations,
    visualise_xsection_cut_on_original_layout,
)

from kqcircuits.pya_resolver import pya
from kqcircuits.simulations.post_process import PostProcess


def get_epr_correction_elmer_solution(**override_args):
    """Returns ElmerCrossSectionSolution for EPR correction simulations with default arguments.
    Optionally, 'override_args' can be used to override any arguments.
    """
    return ElmerCrossSectionSolution(
        **{
            "linear_system_method": "mg",
            "p_element_order": 3,
            "is_axisymmetric": False,
            "mesh_size": {
                "ms_layer_mer&ma_layer_mer": [0.0005, 0.0005, 0.2],
            },
            "integrate_energies": True,
            "run_inductance_sim": False,
            **override_args,
        }
    )


def get_epr_correction_simulations(
    simulations: list[Simulation],
    path: Path,
    correction_cuts: dict[str, dict] | Callable[[Simulation], dict[str, dict]],
    ma_eps_r: float = 8,
    ms_eps_r: float = 11.4,
    sa_eps_r: float = 4,
    ma_thickness: float = 4.8e-3,  # in µm
    ms_thickness: float = 0.3e-3,  # in µm
    sa_thickness: float = 2.4e-3,  # in µm
    ma_bg_eps_r: float = 1,
    ms_bg_eps_r: float = 11.45,
    sa_bg_eps_r: float = 11.45,
    metal_height: float | None = None,
) -> tuple[list[tuple[Simulation, ElmerCrossSectionSolution]], list[PostProcess]]:
    """Helper function to produce EPR correction simulations.

    Args:
        simulations: list of simulation objects
        path: path to simulation folder
        correction_cuts: dictionary or function that returns a dictionary of correction cuts for given simulation.
        Key is the name of cut and values are dicts containing:
        - p1: pya.DPoint indicating the first end of the cut segment
        - p2: pya.DPoint indicating the second end of the cut segment
        - metal_edges: list of dictionaries indicating metal-edge-region locations and orientations in 2D simulation.
            Can contain keywords:
            - x: lateral distance from p1 to mer metal edge,
            - x_reversed: whether gap is closer to p1 than metal, default=False
            - z: height of substrate-vacuum interface, default=0
            - z_reversed: whether vacuum is below substrate, default=False
        - partition_regions: (optional) list of partition region names. The correction cut key is used if not assigned.
        - simulations: (optional) list of simulation names. Is applied on all simulations if not assigned.
        - solution: (optional) solution object for the sim. `get_epr_correction_elmer_solution` is used if not assigned.
            If `solution` is not set, all items under `correction_cuts[Key]`
            are given to `get_epr_correction_elmer_solution` except
            items with keys `["p1", "p2", "metal_edges", "partition_regions", "simulations"]`.
        ma_eps_r: relative permittivity of MA layer
        ms_eps_r: relative permittivity of MS layer
        sa_eps_r: relative permittivity of SA layer
        ma_thickness: thickness of MA layer
        ms_thickness: thickness of MS layer
        sa_thickness: thickness of SA layer
        ma_bg_eps_r: rel permittivity at the location of MA layer in 3D simulation (sheet approximation in use)
        ms_bg_eps_r: rel permittivity at the location of MS layer in 3D simulation (sheet approximation in use)
        sa_bg_eps_r: rel permittivity at the location of SA layer in 3D simulation (sheet approximation in use)
        metal_height: height of metal layers in correction simulations. Use None to get heights from 3D stack

    Returns:
        tuple containing list of correction simulations and list of post_process objects
    """
    correction_simulations = []
    correction_layout = pya.Layout()
    source_sims = {sim[0] if isinstance(sim, Sequence) else sim for sim in simulations}
    for source_sim in source_sims:
        if metal_height is not None:
            source_sim.metal_height = metal_height
        separate_signal_layer_shapes(source_sim)
        cuts = correction_cuts(source_sim) if callable(correction_cuts) else correction_cuts

        for key, cut in cuts.items():
            if "simulations" in cut and source_sim.name not in cut["simulations"]:
                continue

            part_names = cut.get("partition_regions", [key])
            if not part_names:
                raise ValueError("Correction cut has no partition region attached")

            parts = [p for p in source_sim.get_partition_regions() if p.name in part_names]
            if not parts:
                continue

            v_dims = get_list_of_two(parts[0].vertical_dimensions)
            h_dims = get_list_of_two(parts[0].metal_edge_dimensions)
            if None in v_dims or any(v_dims != get_list_of_two(p.vertical_dimensions) for p in parts):
                raise ValueError(f"Partition region vertical_dimensions are invalid for correction_cut {key}.")
            if None in h_dims or any(h_dims != get_list_of_two(p.metal_edge_dimensions) for p in parts):
                raise ValueError(f"Partition region metal_edge_dimensions are invalid for correction_cut {key}.")

            mer_box = []
            for me in cut["metal_edges"]:
                x = me.get("x", 0)
                z = me.get("z", 0)
                dx = int(me.get("x_reversed", False))
                dz = int(me.get("z_reversed", False))
                mer_box.append(pya.DBox(x - h_dims[1 - dx], z - v_dims[dz], x + h_dims[dx], z + v_dims[1 - dz]))

            if "solution" in cut:
                cut_solution = cut["solution"]
            else:
                cut_solution = get_epr_correction_elmer_solution(
                    **{
                        k: v
                        for k, v in cut.items()
                        if k not in ["p1", "p2", "metal_edges", "partition_regions", "simulations"]
                    }
                )

            cords_list = [(cut["p1"], cut["p2"])]
            correction_simulations += cross_combine(
                create_xsections_from_simulations(
                    [source_sim],
                    path,
                    cords_list,
                    layout=correction_layout,
                    ma_permittivity=ma_eps_r,
                    ms_permittivity=ms_eps_r,
                    sa_permittivity=sa_eps_r,
                    ma_thickness=ma_thickness,
                    ms_thickness=ms_thickness,
                    sa_thickness=sa_thickness,
                    magnification_order=1,
                    process_path=XSECTION_PROCESS_PATH,
                    mer_box=mer_box,
                ),
                cut_solution,
            )
            correction_simulations[-1][0].name = correction_simulations[-1][0].cell.name = source_sim.name + "_" + key
            visualise_xsection_cut_on_original_layout([source_sim], cords_list, cut_label=key, width_ratio=0.03)

    post_process = [
        PostProcess(
            "produce_epr_table.py",
            data_file_prefix="correction",
            sheet_approximations={
                "MA": {"thickness": ma_thickness * 1e-6, "eps_r": ma_eps_r, "background_eps_r": ma_bg_eps_r},
                "MS": {"thickness": ms_thickness * 1e-6, "eps_r": ms_eps_r, "background_eps_r": ms_bg_eps_r},
                "SA": {"thickness": sa_thickness * 1e-6, "eps_r": sa_eps_r, "background_eps_r": sa_bg_eps_r},
            },
            groups=["MA", "MS", "SA", "substrate", "vacuum"],
            region_corrections={
                **{p.name: None for s in source_sims for p in s.get_partition_regions()},
                **{p: k for k, v in cuts.items() for p in v.get("partition_regions", [k])},
            },
        ),
    ]
    return correction_simulations, post_process



# This code is part of KQCircuits
# Copyright (C) 2024 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).

from kqcircuits.pya_resolver import pya
from kqcircuits.simulations.partition_region import PartitionRegion
from kqcircuits.simulations.simulation import Simulation


# Partition region and correction cuts definitions for Swissmon qubit
vertical_dimension = 3.0
metal_edge_dimension = 4.0


def partition_regions(simulation: Simulation, prefix: str = "") -> list[PartitionRegion]:

    base_rf = simulation.refpoints["base"]
    port_a_rf = simulation.refpoints["port_a"]
    port_b_rf = simulation.refpoints["port_b"]

    scale = simulation.finger_width + simulation.finger_gap
    s_len = scale * (2 * simulation.finger_control - 3)  # length of straight segment
    width = scale * simulation.finger_control - simulation.finger_width / 2 - simulation.finger_gap / 2

    box_dp = pya.DPoint(s_len / 2.0 + simulation.finger_width + simulation.finger_gap, width)

    a2 = simulation.a if simulation.a2 < 0 else simulation.a2
    b2 = simulation.b if simulation.b2 < 0 else simulation.b2
    rr = b2 + a2 / 2
    rr = (rr + (simulation.finger_gap + 1.5 * simulation.finger_width)) / 2
    rr /= simulation.layout.dbu

    box = pya.DBox(base_rf - box_dp, base_rf + box_dp)
    box_rounded = pya.Region(box.to_itype(simulation.layout.dbu)).rounded_corners(rr, rr, simulation.n)

    port_a_len = 11 + simulation.waveguide_length + metal_edge_dimension  # 11 is the hardcoded port dimension
    port_a_width = simulation.a + 2 * simulation.b + 2 * metal_edge_dimension
    port_a_middle = port_a_rf - pya.DPoint(port_a_len / 2.0, 0)
    port_a_dp = pya.DPoint(port_a_len / 2.0, port_a_width / 2)
    port_a_region = pya.DBox(port_a_middle - port_a_dp, port_a_middle + port_a_dp)

    port_b_len = 11 + simulation.waveguide_length + metal_edge_dimension  # 11 is the hardcoded port dimension
    port_b_width = a2 + 2 * b2 + 2 * metal_edge_dimension
    port_b_middle = port_b_rf + pya.DPoint(port_b_len / 2.0, 0)
    port_b_dp = pya.DPoint(port_b_len / 2.0, port_b_width / 2)
    port_b_region = pya.DBox(port_b_middle - port_b_dp, port_b_middle + port_b_dp)

    result = [
        PartitionRegion(
            name=f"{prefix}port_bmer",
            face=simulation.face_ids[0],
            metal_edge_dimensions=metal_edge_dimension,
            region=port_b_region,
            vertical_dimensions=vertical_dimension,
            visualise=True,
        ),
        PartitionRegion(
            name=f"{prefix}port_bbulk",
            face=simulation.face_ids[0],
            metal_edge_dimensions=None,
            region=port_b_region,
            vertical_dimensions=vertical_dimension,
            visualise=True,
        ),
        PartitionRegion(
            name=f"{prefix}port_amer",
            face=simulation.face_ids[0],
            metal_edge_dimensions=metal_edge_dimension,
            region=port_a_region,
            vertical_dimensions=vertical_dimension,
            visualise=True,
        ),
        PartitionRegion(
            name=f"{prefix}port_abulk",
            face=simulation.face_ids[0],
            metal_edge_dimensions=None,
            region=port_a_region,
            vertical_dimensions=vertical_dimension,
            visualise=True,
        ),
        PartitionRegion(
            name=f"{prefix}fingersmer",
            face=simulation.face_ids[0],
            metal_edge_dimensions=metal_edge_dimension,
            region=box_rounded,
            vertical_dimensions=vertical_dimension,
            visualise=True,
        ),
        PartitionRegion(
            name=f"{prefix}fingersbulk",
            face=simulation.face_ids[0],
            metal_edge_dimensions=None,
            region=box_rounded,
            vertical_dimensions=vertical_dimension,
            visualise=True,
        ),
        PartitionRegion(
            name=f"{prefix}bcomplementmer",
            face=simulation.face_ids[0],
            metal_edge_dimensions=metal_edge_dimension,
            region=None,
            vertical_dimensions=vertical_dimension,
            visualise=True,
        ),
        PartitionRegion(
            name=f"{prefix}bcomplementbulk",
            face=simulation.face_ids[0],
            region=None,
            vertical_dimensions=vertical_dimension,
            visualise=True,
        ),
    ]

    if len(simulation.face_stack) > 1:
        result += [
            PartitionRegion(
                name=f"{prefix}tcomplementmer",
                face=simulation.face_ids[1],
                metal_edge_dimensions=metal_edge_dimension,
                region=None,
                vertical_dimensions=vertical_dimension,
                visualise=True,
            ),
            PartitionRegion(
                name=f"{prefix}tcomplementbulk",
                face=simulation.face_ids[1],
                region=None,
                vertical_dimensions=vertical_dimension,
                visualise=True,
            ),
        ]

    return result


def correction_cuts(simulation: Simulation, prefix: str = "") -> dict[str, dict]:
    base_rf = simulation.refpoints["base"]
    port_a_rf = simulation.refpoints["port_a"]
    port_b_rf = simulation.refpoints["port_b"]

    a2 = simulation.a if simulation.a2 < 0 else simulation.a2
    b2 = simulation.b if simulation.b2 < 0 else simulation.b2

    port_a_len = 11 + simulation.waveguide_length + metal_edge_dimension  # 11 is the hardcoded port dimension
    port_a_middle = port_a_rf - pya.DPoint(port_a_len / 2.0, 0)
    port_a_width = simulation.a + 2 * simulation.b + 2 * metal_edge_dimension

    port_b_len = 11 + simulation.waveguide_length + metal_edge_dimension  # 11 is the hardcoded port dimension
    port_b_middle = port_b_rf + pya.DPoint(port_b_len / 2.0, 0)
    port_b_width = a2 + 2 * b2 + 2 * metal_edge_dimension

    if len(simulation.face_stack) == 1:
        z_me = 0
    else:
        z_me = -simulation.substrate_height[1] - simulation.chip_distance - 2 * simulation.metal_height

    scale = simulation.finger_width + simulation.finger_gap
    half_cut_len = 25.0
    result = {
        f"{prefix}fingersmer": {
            "p1": base_rf + pya.DPoint(0, -scale / 2),
            "p2": base_rf + pya.DPoint(0, scale / 2),
            "metal_edges": [
                {"x": (scale - simulation.finger_gap) / 2, "z": z_me},
                {"x": (scale + simulation.finger_gap) / 2, "z": z_me},
            ],
        },
        f"{prefix}port_amer": {
            "p1": port_a_middle - pya.DPoint(0, port_a_width),
            "p2": port_a_middle + pya.DPoint(0, port_a_width),
            "metal_edges": [
                {
                    "x": port_a_width - simulation.a / 2.0 - simulation.b,
                    "z": z_me,
                },
                {
                    "x": port_a_width - simulation.a / 2.0,
                    "z": z_me,
                },
                {
                    "x": port_a_width + simulation.a / 2.0,
                    "z": z_me,
                },
                {
                    "x": port_a_width + simulation.a / 2.0 + simulation.b,
                    "z": z_me,
                },
            ],
        },
        f"{prefix}port_bmer": {
            "p1": port_b_middle - pya.DPoint(0, port_b_width),
            "p2": port_b_middle + pya.DPoint(0, port_b_width),
            "metal_edges": [
                {"x": port_b_width - a2 / 2 - b2, "z": z_me},
                {"x": port_b_width - a2 / 2, "z": z_me},
                {"x": port_b_width + a2 / 2, "z": z_me},
                {"x": port_b_width + a2 / 2 + b2, "z": z_me},
            ],
        },
        f"{prefix}bcomplementmer": {
            "p1": port_a_rf + pya.DPoint(-half_cut_len, simulation.a + simulation.b),
            "p2": port_a_rf + pya.DPoint(half_cut_len, simulation.a + simulation.b),
            "metal_edges": [
                {"x": half_cut_len, "z": z_me},
            ],
        },
    }
    return result



# This code is part of KQCircuits
# Copyright (C) 2024 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).

from typing import Callable
from kqcircuits.pya_resolver import pya
from kqcircuits.simulations.epr.utils import extract_child_simulation
from kqcircuits.simulations.partition_region import PartitionRegion
from kqcircuits.simulations.simulation import Simulation

# This is an example Python file that showcases the format of how to define
# partition regions and correction cuts for an individual element (e.g. qubit)
# to calculate EPR.


def partition_regions(simulation: Simulation, prefix: str = "") -> list[PartitionRegion]:
    """Derives partition regions needed to simulate a single element.
    Function pointer can be either passed to ``get_single_element_sim_class``
    as ``partition_region_function`` argument,
    or a manually implemented ``Simulation`` class could implement
    the ``get_partition_regions`` function such that it calls this function.

    Args:
        simulation: refpoints and parameters of the single element simulation
            are used to derive the partition regions
        prefix: optional prefix differentiates sets of partition regions
            if extracted from multiple elements within ``simulation``

    Returns:
        List of partition regions
    """
    return [
        PartitionRegion(
            name=f"{prefix}1region",
            face=simulation.face_ids[0],
            metal_edge_dimensions=4.0,
            region=pya.DBox(
                simulation.refpoints["base"] - pya.DPoint(simulation.a, simulation.b),
                simulation.refpoints["base"] + pya.DPoint(simulation.a, simulation.b),
            ),
            vertical_dimensions=3.0,
            visualise=True,
        ),
        PartitionRegion(
            name=f"{prefix}2region",
            face=simulation.face_ids[0],
            metal_edge_dimensions=4.0,
            region=pya.DBox(
                simulation.refpoints["coupler"] - pya.DPoint(simulation.a, simulation.b),
                simulation.refpoints["coupler"] + pya.DPoint(simulation.a, simulation.b),
            ),
            vertical_dimensions=3.0,
            visualise=True,
        ),
        PartitionRegion(
            name=f"{prefix}3region",
            face=simulation.face_ids[0],
            metal_edge_dimensions=4.0,
            region=pya.DBox(
                simulation.refpoints["ro"] - pya.DPoint(simulation.a, simulation.b),
                simulation.refpoints["ro"] + pya.DPoint(simulation.a, simulation.b),
            ),
            vertical_dimensions=3.0,
            visualise=True,
        ),
    ]


def correction_cuts(simulation: Simulation, prefix: str = "") -> dict[str, dict]:
    """Derives correction cuts for each partition region in a single element.
    This function is meant to be passed to ``get_epr_correction_simulations``
    as ``correction_cuts`` argument. If a composition of multiple such functions
    is needed (for example ``simulation`` contains multiple elements, for each
    we want to calculate EPR), a function can be defined that returns a dict
    consisting of multiple ``correction_cuts`` calls, then that function
    can be passed to ``get_epr_correction_simulations``.

    Args:
        simulation: refpoints and parameters of the single element simulation
            are used to derive the partition regions
        prefix: optional prefix differentiates sets of partition regions
            if extracted from multiple elements within ``simulation``

    Returns:
        Configuration dict of correction cuts used by ``get_epr_correction_simulations``
    """
    return {
        f"{prefix}1region": {
            "p1": simulation.refpoints["base"] + pya.DPoint(0, -30),
            "p2": simulation.refpoints["base"] + pya.DPoint(0, 30),
            "metal_edges": [{"x": 20}],
        },
        f"{prefix}2region": {
            "p1": simulation.refpoints["coupler"] + pya.DPoint(simulation.a, -30),
            "p2": simulation.refpoints["coupler"] + pya.DPoint(simulation.a, 30),
            "metal_edges": [{"x": 30}],
        },
        f"{prefix}3region": {
            "p1": simulation.refpoints["ro"] + pya.DPoint(-30, simulation.b),
            "p2": simulation.refpoints["ro"] + pya.DPoint(30, simulation.b),
            "metal_edges": [{"x": 30}],
        },
    }


def extract_from(
    simulation: Simulation, refpoint_prefix: str, parameter_remap_function: Callable[[Simulation, str], any]
):
    """To enable reusing ``partition_regions`` and ``correction_cuts`` functions
    for ``simulation`` that is composed of multiple single elements,
    implement ``extract_from`` function like this, where in the last
    list argument all ``simulation`` parameters (note: not refpoints) that were used
    to implement ``partition_regions`` and ``correction_cuts`` are listed.
    """
    return extract_child_simulation(
        simulation,
        refpoint_prefix,
        parameter_remap_function,
        [
            "face_ids",  # Accesses list's index 0
            "a",
            "b",
        ],
    )
